use ads1256::{
    AdControl, Ads1256, BlockingDelay, DataRate, Multiplexer, OutputPin, SpiBus, Status, Wait,
};
use core::ops::DerefMut;
use embassy_sync::blocking_mutex::raw::RawMutex;
use embassy_sync::mutex::Mutex;
use node_poll_variants::PollVariants;
use physical_node::transducer::input::Poll;
use uom::si::f32;

#[derive(PollVariants)]
#[value_type = "f32::ElectricPotential"]
pub struct AutocalPoll<'a, DeviceMutexT, ModInT, DelayerT, SST, DrdyT, SpiT>
where
    DeviceMutexT: RawMutex,
    ModInT: ModInput,
    DelayerT: BlockingDelay,
    SST: OutputPin,
    DrdyT: Wait,
    SpiT: SpiBus,
{
    input_mod: ModInT,
    ads1256: &'a Mutex<DeviceMutexT, Ads1256<DelayerT, SST, DrdyT>>,
    spi: &'a Mutex<DeviceMutexT, SpiT>,
}

impl<'a, DeviceMutexT, ModInT, DelayerT, SST, DrdyT, SpiT> Poll
    for AutocalPoll<'a, DeviceMutexT, ModInT, DelayerT, SST, DrdyT, SpiT>
where
    DeviceMutexT: RawMutex,
    ModInT: ModInput,
    DelayerT: BlockingDelay,
    SST: OutputPin,
    DrdyT: Wait,
    SpiT: SpiBus,
{
    type Value = f32::ElectricPotential;

    async fn poll(&self) -> Self::Value {
        self.ads1256
            .lock()
            .await
            .deref_mut()
            .autocal_convert_m(
                self.spi,
                self.input_mod.multiplexer(),
                self.input_mod.status(),
                self.input_mod.ad_control(),
                self.input_mod.data_rate(),
                true,
            )
            .await
    }
}

pub trait ModInput {
    fn multiplexer(self) -> Multiplexer;

    fn status(self) -> Option<Status>;

    fn ad_control(self) -> Option<AdControl>;

    fn data_rate(self) -> Option<DataRate>;
}

/// buffer
pub struct ModInStatus {
    pub multiplexer: Multiplexer,
    pub status: Status,
}

impl ModInput for ModInStatus {
    #[inline(always)]
    fn multiplexer(self) -> Multiplexer {
        self.multiplexer
    }

    #[inline(always)]
    fn status(self) -> Option<Status> {
        Some(self.status)
    }

    #[inline(always)]
    fn ad_control(self) -> Option<AdControl> {
        None
    }

    #[inline(always)]
    fn data_rate(self) -> Option<DataRate> {
        None
    }
}

/// gain
pub struct ModInAdControl {
    pub multiplexer: Multiplexer,
    pub ad_control: AdControl,
}

impl ModInput for ModInAdControl {
    #[inline(always)]
    fn multiplexer(self) -> Multiplexer {
        self.multiplexer
    }

    #[inline(always)]
    fn status(self) -> Option<Status> {
        None
    }

    #[inline(always)]
    fn ad_control(self) -> Option<AdControl> {
        Some(self.ad_control)
    }

    #[inline(always)]
    fn data_rate(self) -> Option<DataRate> {
        None
    }
}

/// data rate
pub struct ModInDataRate {
    pub multiplexer: Multiplexer,
    pub data_rate: DataRate,
}

impl ModInput for ModInDataRate {
    #[inline(always)]
    fn multiplexer(self) -> Multiplexer {
        self.multiplexer
    }

    #[inline(always)]
    fn status(self) -> Option<Status> {
        None
    }

    #[inline(always)]
    fn ad_control(self) -> Option<AdControl> {
        None
    }

    #[inline(always)]
    fn data_rate(self) -> Option<DataRate> {
        Some(self.data_rate)
    }
}

/// buffer, gain
pub struct ModInStatAdc {
    pub multiplexer: Multiplexer,
    pub status: Status,
    pub ad_control: AdControl,
}

impl ModInput for ModInStatAdc {
    #[inline(always)]
    fn multiplexer(self) -> Multiplexer {
        self.multiplexer
    }

    #[inline(always)]
    fn status(self) -> Option<Status> {
        Some(self.status)
    }

    #[inline(always)]
    fn ad_control(self) -> Option<AdControl> {
        Some(self.ad_control)
    }

    #[inline(always)]
    fn data_rate(self) -> Option<DataRate> {
        None
    }
}

/// buffer, data rate
pub struct ModInStatDrate {
    pub multiplexer: Multiplexer,
    pub status: Status,
    pub data_rate: DataRate,
}

impl ModInput for ModInStatDrate {
    #[inline(always)]
    fn multiplexer(self) -> Multiplexer {
        self.multiplexer
    }

    #[inline(always)]
    fn status(self) -> Option<Status> {
        Some(self.status)
    }

    #[inline(always)]
    fn ad_control(self) -> Option<AdControl> {
        None
    }

    #[inline(always)]
    fn data_rate(self) -> Option<DataRate> {
        Some(self.data_rate)
    }
}

// gain, data rate
pub struct ModInAdcDrate {
    pub multiplexer: Multiplexer,
    pub ad_control: AdControl,
    pub data_rate: DataRate,
}

impl ModInput for ModInAdcDrate {
    #[inline(always)]
    fn multiplexer(self) -> Multiplexer {
        self.multiplexer
    }

    #[inline(always)]
    fn status(self) -> Option<Status> {
        None
    }

    #[inline(always)]
    fn ad_control(self) -> Option<AdControl> {
        Some(self.ad_control)
    }

    #[inline(always)]
    fn data_rate(self) -> Option<DataRate> {
        Some(self.data_rate)
    }
}

/// buffer, gain, data rate
pub struct ModInAll {
    pub multiplexer: Multiplexer,
    pub status: Status,
    pub ad_control: AdControl,
    pub data_rate: DataRate,
}

impl ModInput for ModInAll {
    #[inline(always)]
    fn multiplexer(self) -> Multiplexer {
        self.multiplexer
    }

    #[inline(always)]
    fn status(self) -> Option<Status> {
        Some(self.status)
    }

    #[inline(always)]
    fn ad_control(self) -> Option<AdControl> {
        Some(self.ad_control)
    }

    #[inline(always)]
    fn data_rate(self) -> Option<DataRate> {
        Some(self.data_rate)
    }
}
